%name C99_STD_GenScanner
%define INHERIT : public Scanner
%define MEMBERS public: virtual void init(void) { init_scanner(); yy_initialize(); }
%define IOSTREAM
%define LEX_PARAM YY_C99_STD_GenParser_STYPE *val, YY_C99_STD_GenParser_LTYPE *loc

%header{
#include <sstream>
#include "Scanner.h"
#include "C99_STD_GenParser.h"

#define LOG(t)		do { count(t); } while (0)
%}

%x COMMENT

IS				(([uU]([lL]|"ll"|"LL")?)|([lL][uU]?)|(("ll"|"LL")[uU]?))
D				[0-9]
D0				[1-9]
A				[_a-zA-Z]
AD				[_a-zA-Z0-9]
H				[0-9a-fA-F]
O				[0-7]

%%

"/*"	{
		LOG(yytext);
		BEGIN COMMENT;
	}
"//".*\n	{
		++line;
		column = 1;
	}

<COMMENT>"*/"	{
		LOG(yytext);
		BEGIN 0;
	}
<COMMENT>\n	{
		++line;
		column = 1;
	}
<COMMENT>.	{
		++column;
	}

"auto"	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::AUTO;
	}
"break"	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::BREAK;
	}
"case"	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::CASE;
	}
"char"	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::CHAR;
	}
"const"	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::CONST;
	}
"continue"	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::CONTINUE;
	}
"default"	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::DEFAULT;
	}
"do"	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::DO;
	}
"double"	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::DOUBLE;
	}
"else"	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::ELSE;
	}
"enum"	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::ENUM;
	}
"extern"	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::EXTERN;
	}
"float"	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::FLOAT;
	}
"for"	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::FOR;
	}
"goto"	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::GOTO;
	}
"if"	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::IF;
	}
"inline"	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::INLINE;
	}
"int"	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::INT;
	}
"long"	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::LONG;
	}
"register"	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::REGISTER;
	}
"restrict"	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::RESTRICT;
	}
"return"	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::RETURN;
	}
"short"		{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::SHORT;
	}
"signed"	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::SIGNED;
	}
"sizeof"	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::SIZEOF;
	}
"static"	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::STATIC;
	}
"struct"	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::STRUCT;
	}
"switch"	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::SWITCH;
	}
"typedef"	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::TYPEDEF;
	}
"union"		{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::UNION;
	}
"unsigned"	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::UNSIGNED;
	}
"void"	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::VOID;
	}
"volatile"	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::VOLATILE;
	}
"while"	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::WHILE;
	}
"_Bool"	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::_BOOL;
	}
"_Complex"	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::_COMPLEX;
	}
"_Imaginary"	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::_IMAGINARY;
	}
"..."	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::ELLIPSIS; 
	}
"<<="	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::LS_ASSIGN;
	}
">>="	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::RS_ASSIGN;
	}
"->"	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::DART;
	}
"=="	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::EQ_COMPARE;
	}
"!="	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::NE_COMPARE;
	}
">="	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::GE_COMPARE;
	}
"<="	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::LE_COMPARE;
	}
"~="	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::NEG_ASSIGN;
	}
"+="	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::ADD_ASSIGN;
	}
"*="	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::MUL_ASSIGN;
	}
"/="	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::DIV_ASSIGN;
	}
"%="	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::MOD_ASSIGN;
	}
"-="	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::SUB_ASSIGN;
	}
"^="	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::XOR_ASSIGN;
	}
"&="	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::AND_ASSIGN;
	}
"|="	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::OR_ASSIGN;
	}
"++"	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::INCREMENT;
	}
"--"	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::DECREMENT;
	}
"&&"	{ 
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::LOGICAL_AND;
	}
"||"	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::LOGICAL_OR;
	}
"<<"	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::LS;
	}
">>"	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::RS;
	}
"="	{
		LOG(yytext);
		memset(val->identifier, 0, SYNTAX_TOKEN_SIZE);
		val->character = yytext[0];
		return C99_STD_GenParser::ASSIGN;
	}
";"	{
		LOG(yytext);
		memset(val->identifier, 0, SYNTAX_TOKEN_SIZE);
		val->character = yytext[0];
		return C99_STD_GenParser::SEMICOLON;
	}
":"	{
		LOG(yytext);
		memset(val->identifier, 0, SYNTAX_TOKEN_SIZE);
		val->character = yytext[0];
		return C99_STD_GenParser::COLON;
	}
","	{
		LOG(yytext);
		memset(val->identifier, 0, SYNTAX_TOKEN_SIZE);
		val->character = yytext[0];
		return C99_STD_GenParser::COMMA;
	}
"."	{
		LOG(yytext);
		memset(val->identifier, 0, SYNTAX_TOKEN_SIZE);
		val->character = yytext[0];
		return C99_STD_GenParser::POINT;
	}
"&"	{
		LOG(yytext);
		memset(val->identifier, 0, SYNTAX_TOKEN_SIZE);
		val->character = yytext[0];
		return C99_STD_GenParser::AND;
	}
"|"	{
		LOG(yytext);
		memset(val->identifier, 0, SYNTAX_TOKEN_SIZE);
		val->character = yytext[0];
		return C99_STD_GenParser::OR;
	}
"~"	{
		LOG(yytext);
		memset(val->identifier, 0, SYNTAX_TOKEN_SIZE);
		val->character = yytext[0];
		return C99_STD_GenParser::NEG;
	}
"^"	{
		LOG(yytext);
		memset(val->identifier, 0, SYNTAX_TOKEN_SIZE);
		val->character = yytext[0];
		return C99_STD_GenParser::XOR;
	}
"!"	{
		LOG(yytext);
		memset(val->identifier, 0, SYNTAX_TOKEN_SIZE);
		val->character = yytext[0];
		return C99_STD_GenParser::NOT;
	}
"\""	{
		LOG(yytext);
		memset(val->identifier, 0, SYNTAX_TOKEN_SIZE);
		val->character = yytext[0];
		return '"';
	}
"'"	{
		LOG(yytext);
		memset(val->identifier, 0, SYNTAX_TOKEN_SIZE);
		val->character = yytext[0];
		return C99_STD_GenParser::QUOTE; }
"+"	{
		LOG(yytext);
		memset(val->identifier, 0, SYNTAX_TOKEN_SIZE);
		val->character = yytext[0];
		return C99_STD_GenParser::PLUS; }
"*"	{
		LOG(yytext);
		memset(val->identifier, 0, SYNTAX_TOKEN_SIZE);
		val->character = yytext[0];
		return C99_STD_GenParser::STAR; }
"-"	{
		LOG(yytext);
		memset(val->identifier, 0, SYNTAX_TOKEN_SIZE);
		val->character = yytext[0];
		return C99_STD_GenParser::MINUS; }
"/"	{
		LOG(yytext);
		memset(val->identifier, 0, SYNTAX_TOKEN_SIZE);
		val->character = yytext[0];
		return C99_STD_GenParser::SLASH; }
"%"	{
		LOG(yytext);
		memset(val->identifier, 0, SYNTAX_TOKEN_SIZE);
		val->character = yytext[0];
		return C99_STD_GenParser::MOD; }
"?"	{
		LOG(yytext);
		memset(val->identifier, 0, SYNTAX_TOKEN_SIZE);
		val->character = yytext[0];
		return C99_STD_GenParser::QUESTION; }
"\\"	{
		LOG(yytext);
		memset(val->identifier, 0, SYNTAX_TOKEN_SIZE);
		val->character = yytext[0];
		return C99_STD_GenParser::BACKSLASH;
	}
"["|"<:"	{
		LOG(yytext);
		memset(val->identifier, 0, SYNTAX_TOKEN_SIZE);
		val->character = yytext[0];
		return '[';
	}
"]"|":>"	{
		LOG(yytext);
		memset(val->identifier, 0, SYNTAX_TOKEN_SIZE);
		val->character = yytext[0];
		return ']';
	}
"{"|"<%"	{
		LOG(yytext);
		memset(val->identifier, 0, SYNTAX_TOKEN_SIZE);
		val->character = yytext[0];
		enter_scope();
		return '{';
	}
"}"|"%>"	{
		LOG(yytext);
		memset(val->identifier, 0, SYNTAX_TOKEN_SIZE);
		val->character = yytext[0];
		leave_scope();
		return '}';
	}
"("	{
		LOG(yytext);
		memset(val->identifier, 0, SYNTAX_TOKEN_SIZE);
		val->character = yytext[0];
		return '(';
	}
")"	{
		LOG(yytext);
		memset(val->identifier, 0, SYNTAX_TOKEN_SIZE);
		val->character = yytext[0];
		return ')';
	}
"<"	{
		LOG(yytext);
		memset(val->identifier, 0, SYNTAX_TOKEN_SIZE);
		val->character = yytext[0];
		return '<';
	}
">"	{
		LOG(yytext);
		memset(val->identifier, 0, SYNTAX_TOKEN_SIZE);
		val->character = yytext[0];
		return '>';
	}
{A}{AD}*	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return check_type(yytext);
	}
(({D0}{D}*)|0){IS}?	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::DECIMAL_CONSTANT;
	}
0[xX]({H})+{IS}?	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::HEXADECIMAL_CONSTANT;
	}
0({O})+{IS}?	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::OCTAL_CONSTANT;
	}
\\[uU]{H}{4}	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::UNIVERSAL_CHARACTER_NAME;
	}
\\[uU]{H}{8}	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::UNIVERSAL_CHARACTER_NAME;
	}
(({D}*\.{D}+)|({D}+\.{D}*))([eE][+-]?{D}*)?[fFlL]?	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::DECIMAL_FLOAT_CONSTANT;
	}
{D}+[eE][+-]?{D}+[fFlL]?	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::DECIMAL_FLOAT_CONSTANT;
	}
0[xX]{H}+[pP][+-]?{D}+[fFlL]?	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::HEXADECIMAL_FLOAT_CONSTANT;
	}
0[xX](({H}*\.{H}+)|({H}+\.))[pP][+-]?{D}+[fFlL]?	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::HEXADECIMAL_FLOAT_CONSTANT;
	}
[L]?\"([^"\n]|"\\\"")*\"	{
		LOG(yytext);
		copy(val->identifier, yytext, SYNTAX_TOKEN_SIZE);
		return C99_STD_GenParser::STRING;
	}
[L]?'[^'\n]+'	{
		LOG(yytext);
		return C99_STD_GenParser::CHARACTER_SEQUENCE;
	}

#.*\n	{
		// TODO: recognize formats
		// TODO: replace scanf
		// format: [# lineno "module.*"]
		// format: [# lineno "<cmd>"]
		int line_no;
		char module[SYNTAX_MODULE_NAME_SIZE];
		LOG(yytext);
		memset(module, 0, SYNTAX_MODULE_NAME_SIZE);
		sscanf(reinterpret_cast<const char *>(yytext), "# %d %s", &line_no, module);
		set_current_module(module);
		++line;
		column = 1;
	}
\r	{
		LOG(yytext);
	}
\n	{
		++line;
		column = 1;
	}
\t	{
		if (column % tabstop) column += tabstop;
		else column += ((column + tabstop) % tabstop);
	}
" "	{
		LOG(yytext);
	}
\f	{
		LOG(yytext);
		std::cerr << "ERROR: \\f encountered in module '" << get_current_module() << "' on line " << line << std::endl;
	}
<<EOF>>	{
		yyterminate();
	}
.	{
		LOG(yytext);
		std::cerr << "UNKNOWN TOKEN: [" << yytext << " at " << line << ":" << column << std::endl;
	}

%%
