#ifndef __DECLARATION_H__
#define __DECLARATION_H__

#include "stringutils.h"
#include <iostream>

class Declaration
{
	public:
		class Compare
		{
			public:
				inline bool operator() (const Declaration & f1, const Declaration & f2)
					{ return f1.s < f2.s; }
		};
	private:
		std::string name;
		std::string module;
		std::string s;
	private:
		inline void update_s(void)
			{ s = module+":"+name; }
	public:
		Declaration()
			: name(""), module("")
			{ update_s(); }
		Declaration(const Declaration & func)
			: name(func.name), module(func.module)
			{ update_s(); }
		Declaration(const std::string & name, const std::string & module)
			: name(name), module(module)
			{ update_s(); }
		inline std::string str() const
			{ return s; }
		inline bool operator== (const Declaration & other) const
			{ return s == other.s; }
		friend std::ostream & operator << (std::ostream & os, const Declaration & func)
			{ return os << func.s; }
		friend std::istream & operator >> (std::istream & is, Declaration & func)
			{
				std::string s;
				is >> s;
				parse(remove_quotes(s), func);
				return is;
			}

		inline const std::string & get_module(void) const
			{ return module; }
		inline const std::string & get_name(void) const
			{ return name; }
		inline void set_module(const std::string & module)
			{
				this->module = module;
				update_s();
			}

		inline void set_name(const std::string & name)
			{
				this->name = name;
				update_s();
			}

		// str="module:function", return:false if str has different form
		static bool parse(const std::string & str, Declaration & f)
			{
				size_t delm = str.find(":");
				if (delm == std::string::npos) return false;
				f.name = str.substr(delm+1);
				f.module = str.substr(0, delm);
				f.update_s();
				return true;
			}
};

#endif

