#ifndef __NODE_H__
#define __NODE_H__

#include <vector>
#include <string>

class Node
{
	private:
		typedef std::vector<Node *> NContainer;
	public:
		typedef NContainer::const_iterator const_iterator;
	private:
		int type;
		Node * root;
		std::string module;
		std::string data_identifier;
		char data_character;
		NContainer sub;
		static size_t objects;
	private:
		Node(const Node &);
		Node(int type);
		Node(int type, const char * identifier);
		Node(int type, char character);
		inline void clear()
			{
				while (!sub.empty())
				{
					Node * node = sub.back();
					sub.pop_back();
					delete node;
				}
			}
		size_t count_nodes(const Node *) const;
	public:
		Node(int type, const std::string & module);
		Node(int type, const std::string & identifier, const std::string & module);
		Node(int type, char character, const std::string & module);
		~Node()
			{ clear(); --objects; }
		inline void add_front(Node * child)
			{
				if (!child) return;
				child->module.clear();
				child->module.reserve(0);
				child->root = this;
				//sub.push_front(child); // for deque<>
				sub.insert(sub.begin(), child); // for vector<>
			}
		inline void add_back(Node * child)
			{
				if (!child) return;
				child->module.clear();
				child->module.reserve(0);
				child->root = this;
				sub.push_back(child);
			}
		inline void add_front(int type)
			{ add_front(new Node(type)); }
		inline void add_front(int type, const char * identifier)
			{ add_front(new Node(type, identifier)); }
		inline void add_front(int type, char character)
			{ add_front(new Node(type, character)); }
		inline void add_back(int type)
			{ add_back(new Node(type)); }
		inline void add_back(int type, const char * identifier)
			{ add_back(new Node(type, identifier)); }
		inline void add_back(int type, char character)
			{ add_back(new Node(type, character)); }
		inline bool is_type(int type) const
			{ return this->type == type; }
		inline int get_type() const
			{ return this->type; }
		inline size_t count_nodes() const
			{ return count_nodes(this); }
		inline size_t sub_count() const
			{ return sub.size(); }
		inline size_t size() const
			{ return sub.size(); }
		inline const Node * operator [] (int i) const
			{ return sub[i]; }
		inline const std::string & identifier() const
			{ return data_identifier; }
		inline char character() const
			{ return data_character; }
		inline const std::string & get_module() const
			{ return (!root) ? this->module : root->get_module(); }
		inline const_iterator begin() const
			{ return sub.begin(); }
		inline const_iterator end() const
			{ return sub.end(); }
		inline static size_t get_num_objects()
			{ return objects; }
};

#endif
