#ifndef __PARSER_H__
#define __PARSER_H__

#include <iostream>
#include "SyntaxTree.h"

class Parser
{
	public:
		enum Token
		{	PARSER_NULL_TOKEN = 299
			,TOKEN_AUTO                         =  300
			,TOKEN_BREAK                        =  301
			,TOKEN_CASE                         =  302
			,TOKEN_CHAR                         =  303
			,TOKEN_CONST                        =  304
			,TOKEN_CONTINUE                     =  305
			,TOKEN_DEFAULT                      =  306
			,TOKEN_DO                           =  307
			,TOKEN_DOUBLE                       =  308
			,TOKEN_ELSE                         =  309
			,TOKEN_ENUM                         =  310
			,TOKEN_EXTERN                       =  311
			,TOKEN_FLOAT                        =  312
			,TOKEN_FOR                          =  313
			,TOKEN_GOTO                         =  314
			,TOKEN_IF                           =  315
			,TOKEN_INLINE                       =  316
			,TOKEN_INT                          =  317
			,TOKEN_LONG                         =  318
			,TOKEN_REGISTER                     =  319
			,TOKEN_RESTRICT                     =  320
			,TOKEN_RETURN                       =  321
			,TOKEN_SHORT                        =  322
			,TOKEN_SIGNED                       =  323
			,TOKEN_SIZEOF                       =  324
			,TOKEN_STATIC                       =  325
			,TOKEN_STRUCT                       =  326
			,TOKEN_SWITCH                       =  327
			,TOKEN_TYPEDEF                      =  328
			,TOKEN_UNION                        =  329
			,TOKEN_UNSIGNED                     =  330
			,TOKEN_VOID                         =  331
			,TOKEN_VOLATILE                     =  332
			,TOKEN_WHILE                        =  333
			,TOKEN__BOOL                        =  334
			,TOKEN__COMPLEX                     =  335
			,TOKEN__IMAGINARY                   =  336
			,TOKEN_ELLIPSIS                     =  337
			,TOKEN_DART                         =  338
			,TOKEN_LS_ASSIGN                    =  339
			,TOKEN_RS_ASSIGN                    =  340
			,TOKEN_MUL_ASSIGN                   =  341
			,TOKEN_DIV_ASSIGN                   =  342
			,TOKEN_ADD_ASSIGN                   =  343
			,TOKEN_SUB_ASSIGN                   =  344
			,TOKEN_NEG_ASSIGN                   =  345
			,TOKEN_XOR_ASSIGN                   =  346
			,TOKEN_AND_ASSIGN                   =  347
			,TOKEN_OR_ASSIGN                    =  348
			,TOKEN_MOD_ASSIGN                   =  349
			,TOKEN_EQ_COMPARE                   =  350
			,TOKEN_NE_COMPARE                   =  351
			,TOKEN_LE_COMPARE                   =  352
			,TOKEN_GE_COMPARE                   =  353
			,TOKEN_INCREMENT                    =  354
			,TOKEN_DECREMENT                    =  355
			,TOKEN_LOGICAL_AND                  =  356
			,TOKEN_LOGICAL_OR                   =  357
			,TOKEN_LS                           =  358
			,TOKEN_RS                           =  359
			,TOKEN_IDENTIFIER                   =  360
			,TOKEN_TYPE_NAME                    =  361
			,TOKEN_DECIMAL_CONSTANT             =  362
			,TOKEN_HEXADECIMAL_CONSTANT         =  363
			,TOKEN_OCTAL_CONSTANT               =  364
			,TOKEN_UNIVERSAL_CHARACTER_NAME     =  365
			,TOKEN_DECIMAL_FLOAT_CONSTANT       =  366
			,TOKEN_HEXADECIMAL_FLOAT_CONSTANT   =  367
			,TOKEN_STRING                       =  368
			,TOKEN_CHARACTER_SEQUENCE           =  369
			,TOKEN_NOT                          =  370
			,TOKEN_NEG                          =  371
			,TOKEN_MINUS                        =  372
			,TOKEN_PLUS                         =  373
			,TOKEN_STAR                         =  374
			,TOKEN_AND                          =  375
			,TOKEN_ASSIGN                       =  376
			,TOKEN_SEMICOLON                    =  377
			,TOKEN_COLON                        =  378
			,TOKEN_COMMA                        =  379
			,TOKEN_POINT                        =  380
			,TOKEN_XOR                          =  381
			,TOKEN_OR                           =  382
			,TOKEN_QUOTE                        =  383
			,TOKEN_SLASH                        =  384
			,TOKEN_MOD                          =  385 
			,TOKEN_QUESTION                     =  386
			,TOKEN_BACKSLASH                    =  387
			,TOKEN_PARENTHESIS_LEFT             =  388
			,TOKEN_PARENTHESIS_RIGHT            =  389
			,TOKEN_BRACKET_LEFT                 =  390
			,TOKEN_BRACKET_RIGHT                =  391
			,TOKEN_BRACE_LEFT                   =  392
			,TOKEN_BRACE_RIGHT                  =  393
			,TOKEN_CLAMP_LEFT                   =  394
			,TOKEN_CLAMP_RIGHT                  =  395
			,TOKEN_LABEL                        =  396
			,TOKEN_EXPRESSION                   =  397
			,TOKEN_DIRECT_DECLARATOR            =  398
			,TOKEN_TYPE_SPECIFIER               =  399
			,TOKEN_DECLARATION_SPECIFIERS       =  400
			,TOKEN_PARAMETER_DECLARATION        =  401
			,TOKEN_PARAMETER_LIST               =  402
			,TOKEN_PARAMETER_TYPE_LIST          =  403
			,TOKEN_DECLARATOR                   =  404
			,TOKEN_INIT_DECLARATOR_LIST         =  405
			,TOKEN_DECLARATION                  =  406
			,TOKEN_EXTERNAL_DECLARATION         =  407
			,TOKEN_TRANSLATION_UNIT             =  408
			,TOKEN_INIT_DECLARATOR              =  409
			,TOKEN_BLOCK_ITEM_LIST              =  410
			,TOKEN_BLOCK_ITEM                   =  411
			,TOKEN_FUNCTION_DEFINITION          =  412
			,TOKEN_PRIMARY_EXPRESSION           =  413
			,TOKEN_ARGUMENT_EXPRESSION_LIST     =  414
			,TOKEN_ASSIGNMENT_EXPRESSION        =  415
			,TOKEN_POSTFIX_EXPRESSION           =  416
			,TOKEN_UNARY_EXPRESSION             =  417
			,TOKEN_CAST_EXPRESSION              =  418
			,TOKEN_MULTIPLICATIVE_EXPRESSION    =  419
			,TOKEN_ADDITIVE_EXPRESSION          =  420
			,TOKEN_SHIFT_EXPRESSION             =  421
			,TOKEN_RELATIONAL_EXPRESSION        =  422
			,TOKEN_AND_EXPRESSION               =  423
			,TOKEN_EXCLUSIVE_OR_EXPRESSION      =  424
			,TOKEN_INCLUSIVE_OR_EXPRESSION      =  425
			,TOKEN_LOGICAL_AND_EXPRESSION       =  426
			,TOKEN_LOGICAL_OR_EXPRESSION        =  427
			,TOKEN_CONDITIONAL_EXPRESSION       =  428
			,TOKEN_STORAGE_CLASS_SPECIFIER      =  429
			,TOKEN_STRUCT_OR_UNION_SPECIFIER    =  430
			,TOKEN_STATEMENT                    =  431
			,TOKEN_EXPRESSION_STATEMENT         =  432
			,TOKEN_COMPOUND_STATEMENT           =  433
			,TOKEN_SELECTION_STATEMENT          =  434
			,TOKEN_LABELED_STATEMENT            =  435
			,TOKEN_JUMP_STATEMENT               =  436
			,TOKEN_ITERATION_STATEMENT          =  437
			,TOKEN_POINTER                      =  438
			,TOKEN_CONSTANT                     =  439
			,TOKEN_SPECIFIER_QUALIFIER_LIST     =  440
			,TOKEN_ENUM_SPECIFIER               =  441
			,TOKEN_FUNCTION_SPECIFIER           =  442
			,TOKEN_TYPE_QUALIFIER               =  443
			,TOKEN_TYPE_QUALIFIER_LIST          =  444
			,TOKEN_INITIALIZER_LIST             =  445
			,TOKEN_DESIGNATOR_LIST              =  446
			,TOKEN_IDENTIFIER_LIST              =  447
			,TOKEN_ENUMERATOR                   =  448
			,TOKEN_DECLARATION_LIST             =  449
			,TOKEN_ENUMERATOR_LIST              =  450
			,TOKEN_DIRECT_ABSTRACT_DECLARATOR   =  451
			,TOKEN_STRUCT_DECLARATION           =  452
			,TOKEN_STRUCT_DECLARATION_LIST      =  453
			,TOKEN_STRUCT_DECLARATOR            =  454
			,TOKEN_STRUCT_DECLARATOR_LIST       =  455
			,TOKEN_PARAMETER                    =  456
			,TOKEN_INITIALIZER                  =  457
			,TOKEN_DESIGNATION                  =  458
			,TOKEN_DESIGNATOR                   =  459
			,TOKEN_CONSTANT_EXPRESSION          =  460
			,TOKEN_INTRINSIC_EXPRESSION         =  461
			// "Tasking 6.0 r4" proprietary extension (start)
			,TOKEN__BIT                         =  501
			,TOKEN__BITA                        =  502
			,TOKEN__BITWORD                     =  503
			,TOKEN__SFRBIT                      =  504
			,TOKEN__ESFRBIT                     =  505
			,TOKEN__SFR                         =  506
			,TOKEN__ESFR                        =  507
			,TOKEN__AT                          =  508
			,TOKEN__ATBIT                       =  509
			,TOKEN__INLINE                      =  510
			,TOKEN__USM                         =  511
			,TOKEN__NOUSM                       =  512
			,TOKEN__FAR                         =  513
			,TOKEN__HUGE                        =  514
			,TOKEN__SHUGE                       =  515
			,TOKEN__NEAR                        =  516
			,TOKEN__SYSTEM                      =  517
			,TOKEN__IRAM                        =  518
			,TOKEN__GETBIT                      =  519
			,TOKEN__PUTBIT                      =  520
			,TOKEN__INTERRUPT                   =  521
			,TOKEN__USING                       =  522
			,TOKEN_INTERRUPT                    =  523
			,TOKEN_USING                        =  524
			,TOKEN__ROL                         =  525
			,TOKEN__ROR                         =  526
			,TOKEN__TESTCLEAR                   =  527
			,TOKEN__TESTSET                     =  528
			,TOKEN__BFLD                        =  529
			,TOKEN__INT166                      =  530
			,TOKEN__IDLE                        =  531
			,TOKEN__NOP                         =  532
			,TOKEN__PRIOR                       =  533
			,TOKEN__PWRDN                       =  534
			,TOKEN__SRVWDT                      =  535
			,TOKEN__DISWDT                      =  536
			,TOKEN__EINIT                       =  537
			,TOKEN__ATOMIC                      =  538
			,TOKEN__MUL32                       =  539
			,TOKEN__MULU32                      =  540
			,TOKEN__DIV32                       =  541
			,TOKEN__DIVU32                      =  542
			,TOKEN__MOD32                       =  543
			,TOKEN__MODU32                      =  544
			,TOKEN__PAG                         =  545
			,TOKEN__POF                         =  546
			,TOKEN__SEG                         =  547
			,TOKEN__SOF                         =  548
			,TOKEN__MKFP                        =  549
			,TOKEN__MKHP                        =  550
			,TOKEN__MKSP                        =  551
			,TOKEN_BANK                         =  552
			,TOKEN_SFR                          =  553
			,TOKEN_ESFR                         =  554
			,TOKEN_SFRBIT                       =  555
			,TOKEN_ESFRBIT                      =  556
			,TOKEN_TOK_NEAR                     =  557
			,TOKEN_TOK_FAR                      =  558
			,TOKEN_TOK_HUGE                     =  559
			,TOKEN_TOK_SHUGE                    =  560
			,TOKEN_BIT                          =  561
			// "Tasking 6.0 r4" proprietary extension (end)
		};
		enum Meta
		{
			TOKEN_META_UNKNOWN,
			TOKEN_META_IDENTIFIER,
			TOKEN_META_CHARACTER,
			TOKEN_META_PSEUDO
		};
	private:
		typedef struct
		{
			int type;
			char name[64];
			int meta_type;
		} token_info_t;
	private:
		static const token_info_t token_info_map[];
		std::string parsed_module;
	protected:
		SyntaxTree * tree;
	protected:
		virtual void add_type(const std::string &) = 0;
		virtual void set_current_module(const std::string &) = 0;
		virtual std::string module() const = 0;
		void check_for_type_name(Node * spec, Node * decl);
	public:
		virtual int parse(std::istream &, const std::string &, SyntaxTree *) = 0;
		virtual std::string name() = 0;
		inline std::string token_name(const Node * node) const
			{ return (!node) ? std::string("") : tok_name(node->get_type()); }
		inline int token_meta_type(const Node * node) const
			{ return (!node) ? TOKEN_META_UNKNOWN : tok_meta_type(node->get_type()); }
		inline std::string token_name(int type) const
			{ return tok_name(type); }
		inline int token_meta_type(int type) const
			{ return tok_meta_type(type); }
		inline void set_parsed_module(const std::string & module)
			{ parsed_module = module; }
		static std::string tok_name(const Node *);
		static int tok_meta_type(const Node *);
		static std::string tok_name(int);
		static int tok_meta_type(int);
};

#endif

