#ifndef __SCANNER_H__
#define __SCANNER_H__

#include <map>
#include <string>

class Scanner
{
	private:
		typedef std::multimap<int,std::string,std::less<int> > ISMMap;
	private:
		std::string current_module;
		int scope;
		ISMMap types;
	private:
		void remove_type_above_scope_level(int);
	protected:
		int line;
		int column;
		int tabstop;
	protected:
		inline bool contains_type(const std::string & type) const
			{
				for (ISMMap::const_reverse_iterator it = types.rbegin(); it != types.rend(); ++it)
					if (it->second == type) return true;
				return false;
			}
		void init_scanner();
		inline void count(const char * t)
			{ while (*t++) ++column; }
		inline void count(const unsigned char * t)
			{ while (*t++) ++column; }
		inline void copy(char * dest, const char * source, size_t len)
			{ strncpy(dest, source, len); }
		inline void copy(char * dest, const unsigned char * source, size_t len)
			{ strncpy(dest, reinterpret_cast<const char *>(source), len); }
		inline int enter_scope(void)
			{ return ++scope; }
		inline int leave_scope(void)
			{ remove_type_above_scope_level(scope); return --scope; }
	public:
		Scanner();
		virtual ~Scanner();
		inline void add_type(const std::string & type)
			{ types.insert(ISMMap::value_type(scope, type)); }
		inline int check_type(unsigned char * type) const
			{ return check_type(std::string(reinterpret_cast<const char *>(type))); }
		virtual int check_type(const std::string &) const = 0;
		inline int get_line() const
			{ return line; }
		inline int get_column() const
			{ return column; }

		inline std::string get_current_module(void) const
			{ return current_module; }
		void set_current_module(const std::string &);
};

#endif
